from typing import Optional

from .player import Player
from ..interfaces import Entity, FightingEntity, Map


class Item(Entity):
    """
    A class for items
    """
    held: bool
    held_by: Optional["Player"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.held = False

    def drop(self, y: int, x: int) -> None:
        """
        The item is dropped from the inventory onto the floor
        """
        if self.held:
            self.held_by.inventory.remove(self)
            self.held = False
            self.held_by = None
        self.map.add_entity(self)
        self.move(y, x)

    def hold(self, player: "Player") -> None:
        """
        The item is taken from the floor and put into the inventory
        """
        self.held = True
        self.held_by = player
        self.map.remove_entity(self)
        player.inventory.append(self)


class Heart(Item):
    """
    A heart item to return health to the player
    """
    name: str = "heart"
    healing: int = 5

    def hold(self, player: "Player") -> None:
        """
        When holding a heart, heal the player and don't put item in inventory.
        """
        player.health = min(player.maxhealth, player.health + self.healing)
        self.map.remove_entity(self)


class Bomb(Item):
    """
    A bomb item intended to deal damage to ennemies at long range
    """
    name: str = "bomb"
    damage: int = 5
    exploding: bool

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.exploding = False

    def drop(self, x: int, y: int) -> None:
        super().drop(x, y)
        self.exploding = True

    def act(self, m: Map) -> None:
        """
        Special exploding action of the bomb
        """
        if self.exploding:
            for e in m.entities:
                if abs(e.x - self.x) + abs(e.y - self.y) <= 1 and \
                        isinstance(e, FightingEntity):
                    e.take_damage(self, self.damage)
